@@ -25,20 +25,10 @@
 #include <net/route.h>
 #include <net/xfrm.h>
 
-static inline void skb_clear_rxhash(struct sk_buff *skb)
-{
-#ifdef HAVE_RXHASH
-	skb->rxhash = 0;
-#endif
-#ifdef HAVE_L4_RXHASH
-	skb->l4_rxhash = 0;
-#endif
-}
-
-#ifdef HAVE_PARALLEL_OPS
-#define SET_PARALLEL_OPS	.parallel_ops = true,
+#ifdef HAVE_GENL_MULTICAST_GROUP_WITH_ID
+#define GROUP_ID(grp)	((grp)->id)
 #else
-#define SET_PARALLEL_OPS
+#define GROUP_ID(grp)	0
 #endif
 
 #if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,36)
