@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2008, 2009, 2010, 2011, 2012, 2013 Nicira, Inc.
+ * Copyright (c) 2008, 2009, 2010, 2011, 2012, 2013, 2015 Nicira, Inc.
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -105,6 +105,13 @@ static enum ofperr process_switch_features(struct lswitch *,
 static void process_packet_in(struct lswitch *, const struct ofp_header *);
 static void process_echo_request(struct lswitch *, const struct ofp_header *);
 
+static ofp_port_t get_mac_entry_ofp_port(const struct mac_learning *ml,
+                                         const struct mac_entry *)
+    OVS_REQ_RDLOCK(ml->rwlock);
+static void set_mac_entry_ofp_port(struct mac_learning *ml,
+                                   struct mac_entry *, ofp_port_t)
+    OVS_REQ_WRLOCK(ml->rwlock);
+
 /* Creates and returns a new learning switch whose configuration is given by
  * 'cfg'.
  *
@@ -315,12 +322,12 @@ lswitch_process_packet(struct lswitch *sw, const struct ofpbuf *msg)
 
     switch (type) {
     case OFPTYPE_ECHO_REQUEST:
-        process_echo_request(sw, msg->data);
+        process_echo_request(sw, ofpbuf_data(msg));
         break;
 
     case OFPTYPE_FEATURES_REPLY:
         if (sw->state == S_FEATURES_REPLY) {
-            if (!process_switch_features(sw, msg->data)) {
+            if (!process_switch_features(sw, ofpbuf_data(msg))) {
                 sw->state = S_SWITCHING;
             } else {
                 rconn_disconnect(sw->rconn);
@@ -329,7 +336,7 @@ lswitch_process_packet(struct lswitch *sw, const struct ofpbuf *msg)
         break;
 
     case OFPTYPE_PACKET_IN:
-        process_packet_in(sw, msg->data);
+        process_packet_in(sw, ofpbuf_data(msg));
         break;
 
     case OFPTYPE_FLOW_REMOVED:
@@ -398,9 +405,11 @@ lswitch_process_packet(struct lswitch *sw, const struct ofpbuf *msg)
     case OFPTYPE_METER_FEATURES_STATS_REPLY:
     case OFPTYPE_TABLE_FEATURES_STATS_REQUEST:
     case OFPTYPE_TABLE_FEATURES_STATS_REPLY:
+    case OFPTYPE_BUNDLE_CONTROL:
+    case OFPTYPE_BUNDLE_ADD_MESSAGE:
     default:
         if (VLOG_IS_DBG_ENABLED()) {
-            char *s = ofp_to_string(msg->data, msg->size, 2);
+            char *s = ofp_to_string(ofpbuf_data(msg), ofpbuf_size(msg), 2);
             VLOG_DBG_RL(&rl, "%016llx: OpenFlow packet ignored: %s",
                         sw->datapath_id, s);
             free(s);
@@ -483,14 +492,14 @@ lswitch_choose_destination(struct lswitch *sw, const struct flow *flow)
         if (mac_learning_may_learn(sw->ml, flow->dl_src, 0)) {
             struct mac_entry *mac = mac_learning_insert(sw->ml, flow->dl_src,
                                                         0);
-            if (mac->port.ofp_port != flow->in_port.ofp_port) {
+            if (get_mac_entry_ofp_port(sw->ml, mac)
+                != flow->in_port.ofp_port) {
                 VLOG_DBG_RL(&rl, "%016llx: learned that "ETH_ADDR_FMT" is on "
                             "port %"PRIu16, sw->datapath_id,
                             ETH_ADDR_ARGS(flow->dl_src),
                             flow->in_port.ofp_port);
 
-                mac->port.ofp_port = flow->in_port.ofp_port;
-                mac_learning_changed(sw->ml);
+                set_mac_entry_ofp_port(sw->ml, mac, flow->in_port.ofp_port);
             }
         }
         ovs_rwlock_unlock(&sw->ml->rwlock);
@@ -508,7 +517,7 @@ lswitch_choose_destination(struct lswitch *sw, const struct flow *flow)
         ovs_rwlock_rdlock(&sw->ml->rwlock);
         mac = mac_learning_lookup(sw->ml, flow->dl_dst, 0);
         if (mac) {
-            out_port = mac->port.ofp_port;
+            out_port = get_mac_entry_ofp_port(sw->ml, mac);
             if (out_port == flow->in_port.ofp_port) {
                 /* Don't send a packet back out its input port. */
                 ovs_rwlock_unlock(&sw->ml->rwlock);
@@ -556,7 +565,6 @@ process_packet_in(struct lswitch *sw, const struct ofp_header *oh)
 
     struct ofpbuf pkt;
     struct flow flow;
-    union flow_in_port in_port_;
 
     error = ofputil_decode_packet_in(&pi, oh);
     if (error) {
@@ -574,8 +582,8 @@ process_packet_in(struct lswitch *sw, const struct ofp_header *oh)
 
     /* Extract flow data from 'opi' into 'flow'. */
     ofpbuf_use_const(&pkt, pi.packet, pi.packet_len);
-    in_port_.ofp_port = pi.fmd.in_port;
-    flow_extract(&pkt, 0, 0, NULL, &in_port_, &flow);
+    flow_extract(&pkt, NULL, &flow);
+    flow.in_port.ofp_port = pi.fmd.in_port;
     flow.tunnel.tun_id = pi.fmd.tun_id;
 
     /* Choose output port. */
@@ -599,15 +607,15 @@ process_packet_in(struct lswitch *sw, const struct ofp_header *oh)
     /* Prepare packet_out in case we need one. */
     po.buffer_id = pi.buffer_id;
     if (po.buffer_id == UINT32_MAX) {
-        po.packet = pkt.data;
-        po.packet_len = pkt.size;
+        po.packet = ofpbuf_data(&pkt);
+        po.packet_len = ofpbuf_size(&pkt);
     } else {
         po.packet = NULL;
         po.packet_len = 0;
     }
     po.in_port = pi.fmd.in_port;
-    po.ofpacts = ofpacts.data;
-    po.ofpacts_len = ofpacts.size;
+    po.ofpacts = ofpbuf_data(&ofpacts);
+    po.ofpacts_len = ofpbuf_size(&ofpacts);
 
     /* Send the packet, and possibly the whole flow, to the output port. */
     if (sw->max_idle >= 0 && (!sw->ml || out_port != OFPP_FLOOD)) {
@@ -625,8 +633,8 @@ process_packet_in(struct lswitch *sw, const struct ofp_header *oh)
         fm.idle_timeout = sw->max_idle;
         fm.buffer_id = pi.buffer_id;
         fm.out_port = OFPP_NONE;
-        fm.ofpacts = ofpacts.data;
-        fm.ofpacts_len = ofpacts.size;
+        fm.ofpacts = ofpbuf_data(&ofpacts);
+        fm.ofpacts_len = ofpbuf_size(&ofpacts);
         buffer = ofputil_encode_flow_mod(&fm, sw->protocol);
 
         queue_tx(sw, buffer);
@@ -649,3 +657,20 @@ process_echo_request(struct lswitch *sw, const struct ofp_header *rq)
 {
     queue_tx(sw, make_echo_reply(rq));
 }
+
+static ofp_port_t
+get_mac_entry_ofp_port(const struct mac_learning *ml,
+                       const struct mac_entry *e)
+    OVS_REQ_RDLOCK(ml->rwlock)
+{
+    void *port = mac_entry_get_port(ml, e);
+    return (OVS_FORCE ofp_port_t) (uintptr_t) port;
+}
+
+static void
+set_mac_entry_ofp_port(struct mac_learning *ml,
+                       struct mac_entry *e, ofp_port_t ofp_port)
+    OVS_REQ_WRLOCK(ml->rwlock)
+{
+    mac_entry_set_port(ml, e, (void *) (OVS_FORCE uintptr_t) ofp_port);
+}
