@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2010, 2011, 2012, 2013 Nicira, Inc.
+ * Copyright (c) 2010, 2011, 2012, 2013, 2014 Nicira, Inc.
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -25,7 +25,6 @@
 #include <sys/ioctl.h>
 
 #include "byte-order.h"
-#include "connectivity.h"
 #include "daemon.h"
 #include "dirs.h"
 #include "dpif.h"
@@ -35,8 +34,8 @@
 #include "netdev-provider.h"
 #include "ofpbuf.h"
 #include "packets.h"
+#include "poll-loop.h"
 #include "route-table.h"
-#include "seq.h"
 #include "shash.h"
 #include "socket-util.h"
 #include "vlog.h"
@@ -59,6 +58,8 @@ struct netdev_vport {
 
     /* Tunnels. */
     struct netdev_tunnel_config tnl_cfg;
+    char egress_iface[IFNAMSIZ];
+    bool carrier_status;
 
     /* Patch Ports. */
     char *peer;
@@ -69,9 +70,13 @@ struct vport_class {
     struct netdev_class netdev_class;
 };
 
+/* Last read of the route-table's change number. */
+static uint64_t rt_change_seqno;
+
 static int netdev_vport_construct(struct netdev *);
 static int get_patch_config(const struct netdev *netdev, struct smap *args);
 static int get_tunnel_config(const struct netdev *, struct smap *args);
+static bool tunnel_check_status_change__(struct netdev_vport *);
 
 static bool
 is_vport_class(const struct netdev_class *class)
@@ -79,6 +84,12 @@ is_vport_class(const struct netdev_class *class)
     return class->construct == netdev_vport_construct;
 }
 
+bool
+netdev_vport_is_vport_class(const struct netdev_class *class)
+{
+    return is_vport_class(class);
+}
+
 static const struct vport_class *
 vport_class_cast(const struct netdev_class *class)
 {
@@ -166,6 +177,35 @@ netdev_vport_get_dpif_port_strdup(const struct netdev *netdev)
                                               sizeof namebuf));
 }
 
+/* Whenever the route-table change number is incremented,
+ * netdev_vport_route_changed() should be called to update
+ * the corresponding tunnel interface status. */
+static void
+netdev_vport_route_changed(void)
+{
+    struct netdev **vports;
+    size_t i, n_vports;
+
+    vports = netdev_get_vports(&n_vports);
+    for (i = 0; i < n_vports; i++) {
+        struct netdev *netdev_ = vports[i];
+        struct netdev_vport *netdev = netdev_vport_cast(netdev_);
+
+        ovs_mutex_lock(&netdev->mutex);
+        /* Finds all tunnel vports. */
+        if (netdev->tnl_cfg.ip_dst) {
+            if (tunnel_check_status_change__(netdev)) {
+                netdev_change_seq_changed(netdev_);
+            }
+        }
+        ovs_mutex_unlock(&netdev->mutex);
+
+        netdev_close(netdev_);
+    }
+
+    free(vports);
+}
+
 static struct netdev *
 netdev_vport_alloc(void)
 {
@@ -212,7 +252,7 @@ netdev_vport_set_etheraddr(struct netdev *netdev_,
     ovs_mutex_lock(&netdev->mutex);
     memcpy(netdev->etheraddr, mac, ETH_ADDR_LEN);
     ovs_mutex_unlock(&netdev->mutex);
-    seq_change(connectivity_seq_get());
+    netdev_change_seq_changed(netdev_);
 
     return 0;
 }
@@ -230,27 +270,48 @@ netdev_vport_get_etheraddr(const struct netdev *netdev_,
     return 0;
 }
 
+/* Checks if the tunnel status has changed and returns a boolean.
+ * Updates the tunnel status if it has changed. */
+static bool
+tunnel_check_status_change__(struct netdev_vport *netdev)
+    OVS_REQUIRES(netdev->mutex)
+{
+    char iface[IFNAMSIZ];
+    bool status = false;
+    ovs_be32 route;
+
+    iface[0] = '\0';
+    route = netdev->tnl_cfg.ip_dst;
+    if (route_table_get_name(route, iface)) {
+        struct netdev *egress_netdev;
+
+        if (!netdev_open(iface, "system", &egress_netdev)) {
+            status = netdev_get_carrier(egress_netdev);
+            netdev_close(egress_netdev);
+        }
+    }
+
+    if (strcmp(netdev->egress_iface, iface)
+        || netdev->carrier_status != status) {
+        ovs_strlcpy(netdev->egress_iface, iface, IFNAMSIZ);
+        netdev->carrier_status = status;
+
+        return true;
+    }
+
+    return false;
+}
+
 static int
 tunnel_get_status(const struct netdev *netdev_, struct smap *smap)
 {
     struct netdev_vport *netdev = netdev_vport_cast(netdev_);
-    char iface[IFNAMSIZ];
-    ovs_be32 route;
 
-    ovs_mutex_lock(&netdev->mutex);
-    route = netdev->tnl_cfg.ip_dst;
-    ovs_mutex_unlock(&netdev->mutex);
+    if (netdev->egress_iface[0]) {
+        smap_add(smap, "tunnel_egress_iface", netdev->egress_iface);
 
-    if (route_table_get_name(route, iface)) {
-        struct netdev *egress_netdev;
-
-        smap_add(smap, "tunnel_egress_iface", iface);
-
-        if (!netdev_open(iface, "system", &egress_netdev)) {
-            smap_add(smap, "tunnel_egress_iface_carrier",
-                     netdev_get_carrier(egress_netdev) ? "up" : "down");
-            netdev_close(egress_netdev);
-        }
+        smap_add(smap, "tunnel_egress_iface_carrier",
+                 netdev->carrier_status ? "up" : "down");
     }
 
     return 0;
@@ -273,13 +334,26 @@ netdev_vport_update_flags(struct netdev *netdev OVS_UNUSED,
 static void
 netdev_vport_run(void)
 {
+    uint64_t seq;
+
     route_table_run();
+    seq = route_table_get_change_seq();
+    if (rt_change_seqno != seq) {
+        rt_change_seqno = seq;
+        netdev_vport_route_changed();
+    }
 }
 
 static void
 netdev_vport_wait(void)
 {
+    uint64_t seq;
+
     route_table_wait();
+    seq = route_table_get_change_seq();
+    if (rt_change_seqno != seq) {
+        poll_immediate_wake();
+    }
 }
 
 /* Code specific to tunnel types. */
@@ -433,6 +507,7 @@ set_tunnel_config(struct netdev *dev_, const struct smap *args)
         static struct ovs_mutex mutex = OVS_MUTEX_INITIALIZER;
         static pid_t pid = 0;
 
+#ifndef _WIN32
         ovs_mutex_lock(&mutex);
         if (pid <= 0) {
             char *file_name = xasprintf("%s/%s", ovs_rundir(),
@@ -441,6 +516,7 @@ set_tunnel_config(struct netdev *dev_, const struct smap *args)
             free(file_name);
         }
         ovs_mutex_unlock(&mutex);
+#endif
 
         if (pid < 0) {
             VLOG_ERR("%s: IPsec requires the ovs-monitor-ipsec daemon",
@@ -483,8 +559,11 @@ set_tunnel_config(struct netdev *dev_, const struct smap *args)
                                &tnl_cfg.out_key_flow);
 
     ovs_mutex_lock(&dev->mutex);
-    dev->tnl_cfg = tnl_cfg;
-    seq_change(connectivity_seq_get());
+    if (memcmp(&dev->tnl_cfg, &tnl_cfg, sizeof tnl_cfg)) {
+        dev->tnl_cfg = tnl_cfg;
+        tunnel_check_status_change__(dev);
+        netdev_change_seq_changed(dev_);
+    }
     ovs_mutex_unlock(&dev->mutex);
 
     return 0;
@@ -656,9 +735,11 @@ set_patch_config(struct netdev *dev_, const struct smap *args)
     }
 
     ovs_mutex_lock(&dev->mutex);
-    free(dev->peer);
-    dev->peer = xstrdup(peer);
-    seq_change(connectivity_seq_get());
+    if (!dev->peer || strcmp(dev->peer, peer)) {
+        free(dev->peer);
+        dev->peer = xstrdup(peer);
+        netdev_change_seq_changed(dev_);
+    }
     ovs_mutex_unlock(&dev->mutex);
 
     return 0;
