@@ -1,4 +1,4 @@
-/* Copyright (c) 2009, 2010, 2011, 2012, 2013 Nicira, Inc.
+/* Copyright (c) 2009, 2010, 2011, 2012, 2013, 2014, 2015 Nicira, Inc.
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -27,6 +27,7 @@
 #include "ovsdb-error.h"
 #include "ovsdb-parser.h"
 #include "ovsdb.h"
+#include "poll-loop.h"
 #include "reconnect.h"
 #include "row.h"
 #include "server.h"
@@ -62,6 +63,8 @@ static bool ovsdb_jsonrpc_session_get_status(
     struct ovsdb_jsonrpc_remote_status *);
 static void ovsdb_jsonrpc_session_unlock_all(struct ovsdb_jsonrpc_session *);
 static void ovsdb_jsonrpc_session_unlock__(struct ovsdb_lock_waiter *);
+static void ovsdb_jsonrpc_session_send(struct ovsdb_jsonrpc_session *,
+                                       struct jsonrpc_msg *);
 
 /* Triggers. */
 static void ovsdb_jsonrpc_trigger_create(struct ovsdb_jsonrpc_session *,
@@ -82,8 +85,8 @@ static struct jsonrpc_msg *ovsdb_jsonrpc_monitor_cancel(
     struct json_array *params,
     const struct json *request_id);
 static void ovsdb_jsonrpc_monitor_remove_all(struct ovsdb_jsonrpc_session *);
-static size_t ovsdb_jsonrpc_monitor_json_length_all(
-    struct ovsdb_jsonrpc_session *);
+static void ovsdb_jsonrpc_monitor_flush_all(struct ovsdb_jsonrpc_session *);
+static bool ovsdb_jsonrpc_monitor_needs_flush(struct ovsdb_jsonrpc_session *);
 
 /* JSON-RPC database server. */
 
@@ -195,10 +198,16 @@ ovsdb_jsonrpc_server_set_remotes(struct ovsdb_jsonrpc_server *svr,
     struct shash_node *node, *next;
 
     SHASH_FOR_EACH_SAFE (node, next, &svr->remotes) {
-        if (!shash_find(new_remotes, node->name)) {
+        struct ovsdb_jsonrpc_remote *remote = node->data;
+        struct ovsdb_jsonrpc_options *options
+            = shash_find_data(new_remotes, node->name);
+
+        if (!options) {
             VLOG_INFO("%s: remote deconfigured", node->name);
             ovsdb_jsonrpc_server_del_remote(node);
-        }
+        } else if (options->dscp != remote->dscp) {
+            ovsdb_jsonrpc_server_del_remote(node);
+         }
     }
     SHASH_FOR_EACH (node, new_remotes) {
         const struct ovsdb_jsonrpc_options *options = node->data;
@@ -364,8 +373,6 @@ struct ovsdb_jsonrpc_session {
     struct list node;           /* Element in remote's sessions list. */
     struct ovsdb_session up;
     struct ovsdb_jsonrpc_remote *remote;
-    size_t backlog_threshold;   /* See ovsdb_jsonrpc_session_run(). */
-    size_t reply_backlog;
 
     /* Triggers. */
     struct hmap triggers;       /* Hmap of "struct ovsdb_jsonrpc_trigger"s. */
@@ -383,8 +390,6 @@ static int ovsdb_jsonrpc_session_run(struct ovsdb_jsonrpc_session *);
 static void ovsdb_jsonrpc_session_wait(struct ovsdb_jsonrpc_session *);
 static void ovsdb_jsonrpc_session_get_memory_usage(
     const struct ovsdb_jsonrpc_session *, struct simap *usage);
-static void ovsdb_jsonrpc_session_set_options(
-    struct ovsdb_jsonrpc_session *, const struct ovsdb_jsonrpc_options *);
 static void ovsdb_jsonrpc_session_got_request(struct ovsdb_jsonrpc_session *,
                                              struct jsonrpc_msg *);
 static void ovsdb_jsonrpc_session_got_notify(struct ovsdb_jsonrpc_session *,
@@ -402,8 +407,6 @@ ovsdb_jsonrpc_session_create(struct ovsdb_jsonrpc_remote *remote,
     list_push_back(&remote->sessions, &s->node);
     hmap_init(&s->triggers);
     hmap_init(&s->monitors);
-    s->reply_backlog = 0;
-    s->backlog_threshold = 1024 * 1024;
     s->js = js;
     s->js_seqno = jsonrpc_session_get_seqno(js);
 
@@ -432,8 +435,6 @@ ovsdb_jsonrpc_session_close(struct ovsdb_jsonrpc_session *s)
 static int
 ovsdb_jsonrpc_session_run(struct ovsdb_jsonrpc_session *s)
 {
-    size_t backlog;
-
     jsonrpc_session_run(s->js);
     if (s->js_seqno != jsonrpc_session_get_seqno(s->js)) {
         s->js_seqno = jsonrpc_session_get_seqno(s->js);
@@ -444,9 +445,12 @@ ovsdb_jsonrpc_session_run(struct ovsdb_jsonrpc_session *s)
 
     ovsdb_jsonrpc_trigger_complete_done(s);
 
-    backlog = jsonrpc_session_get_backlog(s->js);
-    if (!backlog) {
-        struct jsonrpc_msg *msg = jsonrpc_session_recv(s->js);
+    if (!jsonrpc_session_get_backlog(s->js)) {
+        struct jsonrpc_msg *msg;
+
+        ovsdb_jsonrpc_monitor_flush_all(s);
+
+        msg = jsonrpc_session_recv(s->js);
         if (msg) {
             if (msg->type == JSONRPC_REQUEST) {
                 ovsdb_jsonrpc_session_got_request(s, msg);
@@ -460,39 +464,6 @@ ovsdb_jsonrpc_session_run(struct ovsdb_jsonrpc_session *s)
                 jsonrpc_msg_destroy(msg);
             }
         }
-        s->reply_backlog = jsonrpc_session_get_backlog(s->js);
-    } else if (backlog > s->reply_backlog + s->backlog_threshold) {
-        /* We have a lot of data queued to send to the client.  The data is
-         * likely to be mostly monitor updates.  It is unlikely that the
-         * monitor updates are due to transactions by 's', because we will not
-         * let 's' make any more transactions until it drains its backlog to 0
-         * (see previous 'if' case).  So the monitor updates are probably due
-         * to transactions made by database clients other than 's'.  We can't
-         * fix that by preventing 's' from executing more transactions.  We
-         * could fix it by preventing every client from executing transactions,
-         * but then one slow or hung client could prevent other clients from
-         * doing useful work.
-         *
-         * Our solution is to cap the maximum backlog to O(1) in the amount of
-         * data in the database.  If the backlog exceeds that amount, then we
-         * disconnect the client.  When it reconnects, it can fetch the entire
-         * contents of the database using less data than was previously
-         * backlogged. */
-        size_t monitor_length;
-
-        monitor_length = ovsdb_jsonrpc_monitor_json_length_all(s);
-        if (backlog > s->reply_backlog + monitor_length * 2) {
-            VLOG_INFO("%s: %"PRIuSIZE" bytes backlogged but a complete replica "
-                      "would only take %"PRIuSIZE" bytes, disconnecting",
-                      jsonrpc_session_get_name(s->js),
-                      backlog - s->reply_backlog, monitor_length);
-            jsonrpc_session_force_reconnect(s->js);
-        } else {
-            /* The backlog is not unreasonably big.  Only check again after it
-             * becomes much bigger. */
-            s->backlog_threshold = 2 * MAX(s->backlog_threshold * 2,
-                                           monitor_length);
-        }
     }
     return jsonrpc_session_is_alive(s->js) ? 0 : ETIMEDOUT;
 }
@@ -524,7 +495,11 @@ ovsdb_jsonrpc_session_wait(struct ovsdb_jsonrpc_session *s)
 {
     jsonrpc_session_wait(s->js);
     if (!jsonrpc_session_get_backlog(s->js)) {
-        jsonrpc_session_recv_wait(s->js);
+        if (ovsdb_jsonrpc_monitor_needs_flush(s)) {
+            poll_immediate_wake();
+        } else {
+            jsonrpc_session_recv_wait(s->js);
+        }
     }
 }
 
@@ -585,7 +560,10 @@ ovsdb_jsonrpc_session_reconnect_all(struct ovsdb_jsonrpc_remote *remote)
 }
 
 /* Sets the options for all of the JSON-RPC sessions managed by 'remote' to
- * 'options'. */
+ * 'options'.
+ *
+ * (The dscp value can't be changed directly; the caller must instead close and
+ * re-open the session.) */
 static void
 ovsdb_jsonrpc_session_set_all_options(
     struct ovsdb_jsonrpc_remote *remote,
@@ -593,22 +571,6 @@ ovsdb_jsonrpc_session_set_all_options(
 {
     struct ovsdb_jsonrpc_session *s;
 
-    if (remote->listener) {
-        int error;
-
-        error = pstream_set_dscp(remote->listener, options->dscp);
-        if (error) {
-            VLOG_ERR("%s: set_dscp failed %s",
-                     pstream_get_name(remote->listener), ovs_strerror(error));
-        } else {
-            remote->dscp = options->dscp;
-        }
-        /*
-         * XXX race window between setting dscp to listening socket
-         * and accepting socket. Accepted socket may have old dscp value.
-         * Ignore this race window for now.
-         */
-    }
     LIST_FOR_EACH (s, node, &remote->sessions) {
         ovsdb_jsonrpc_session_set_options(s, options);
     }
@@ -740,7 +702,7 @@ ovsdb_jsonrpc_session_notify(struct ovsdb_session *session,
 
     s = CONTAINER_OF(session, struct ovsdb_jsonrpc_session, up);
     params = json_array_create_1(json_string_create(lock_name));
-    jsonrpc_session_send(s->js, jsonrpc_create_notify(method, params));
+    ovsdb_jsonrpc_session_send(s, jsonrpc_create_notify(method, params));
 }
 
 static struct jsonrpc_msg *
@@ -915,7 +877,7 @@ ovsdb_jsonrpc_session_got_request(struct ovsdb_jsonrpc_session *s,
 
     if (reply) {
         jsonrpc_msg_destroy(request);
-        jsonrpc_session_send(s->js, reply);
+        ovsdb_jsonrpc_session_send(s, reply);
     }
 }
 
@@ -943,6 +905,14 @@ ovsdb_jsonrpc_session_got_notify(struct ovsdb_jsonrpc_session *s,
     }
     jsonrpc_msg_destroy(request);
 }
+
+static void
+ovsdb_jsonrpc_session_send(struct ovsdb_jsonrpc_session *s,
+                           struct jsonrpc_msg *msg)
+{
+    ovsdb_jsonrpc_monitor_flush_all(s);
+    jsonrpc_session_send(s->js, msg);
+}
 
 /* JSON-RPC database server triggers.
  *
@@ -970,7 +940,7 @@ ovsdb_jsonrpc_trigger_create(struct ovsdb_jsonrpc_session *s, struct ovsdb *db,
 
         msg = jsonrpc_create_error(json_string_create("duplicate request ID"),
                                    id);
-        jsonrpc_session_send(s->js, msg);
+        ovsdb_jsonrpc_session_send(s, msg);
         json_destroy(id);
         json_destroy(params);
         return;
@@ -1021,7 +991,7 @@ ovsdb_jsonrpc_trigger_complete(struct ovsdb_jsonrpc_trigger *t)
             reply = jsonrpc_create_error(json_string_create("canceled"),
                                          t->id);
         }
-        jsonrpc_session_send(s->js, reply);
+        ovsdb_jsonrpc_session_send(s, reply);
     }
 
     json_destroy(t->id);
@@ -1065,6 +1035,14 @@ struct ovsdb_jsonrpc_monitor_column {
     enum ovsdb_jsonrpc_monitor_selection select;
 };
 
+/* A row that has changed in a monitored table. */
+struct ovsdb_jsonrpc_monitor_row {
+    struct hmap_node hmap_node; /* In ovsdb_jsonrpc_monitor_table.changes. */
+    struct uuid uuid;           /* UUID of row that changed. */
+    struct ovsdb_datum *old;    /* Old data, NULL for an inserted row. */
+    struct ovsdb_datum *new;    /* New data, NULL for a deleted row. */
+};
+
 /* A particular table being monitored. */
 struct ovsdb_jsonrpc_monitor_table {
     const struct ovsdb_table *table;
@@ -1076,6 +1054,10 @@ struct ovsdb_jsonrpc_monitor_table {
     /* Columns being monitored. */
     struct ovsdb_jsonrpc_monitor_column *columns;
     size_t n_columns;
+
+    /* Contains 'struct ovsdb_jsonrpc_monitor_row's for rows that have been
+     * updated but not yet flushed to the jsonrpc connection. */
+    struct hmap changes;
 };
 
 /* A collection of tables being monitored. */
@@ -1096,8 +1078,6 @@ struct ovsdb_jsonrpc_monitor *ovsdb_jsonrpc_monitor_find(
 static void ovsdb_jsonrpc_monitor_destroy(struct ovsdb_replica *);
 static struct json *ovsdb_jsonrpc_monitor_get_initial(
     const struct ovsdb_jsonrpc_monitor *);
-static size_t ovsdb_jsonrpc_monitor_json_length(
-    const struct ovsdb_jsonrpc_monitor *);
 
 static bool
 parse_bool(struct ovsdb_parser *parser, const char *name, bool default_value)
@@ -1287,6 +1267,7 @@ ovsdb_jsonrpc_monitor_create(struct ovsdb_jsonrpc_session *s, struct ovsdb *db,
 
         mt = xzalloc(sizeof *mt);
         mt->table = table;
+        hmap_init(&mt->changes);
         shash_add(&m->tables, table->schema->name, mt);
 
         /* Parse columns. */
@@ -1367,22 +1348,6 @@ ovsdb_jsonrpc_monitor_remove_all(struct ovsdb_jsonrpc_session *s)
     }
 }
 
-/* Returns an overestimate of the number of bytes of JSON data required to
- * report the current contents of the database over all the monitors currently
- * configured in 's'.  */
-static size_t
-ovsdb_jsonrpc_monitor_json_length_all(struct ovsdb_jsonrpc_session *s)
-{
-    struct ovsdb_jsonrpc_monitor *m;
-    size_t length;
-
-    length = 0;
-    HMAP_FOR_EACH (m, node, &s->monitors) {
-        length += ovsdb_jsonrpc_monitor_json_length(m);
-    }
-    return length;
-}
-
 static struct ovsdb_jsonrpc_monitor *
 ovsdb_jsonrpc_monitor_cast(struct ovsdb_replica *replica)
 {
@@ -1391,189 +1356,347 @@ ovsdb_jsonrpc_monitor_cast(struct ovsdb_replica *replica)
 }
 
 struct ovsdb_jsonrpc_monitor_aux {
-    bool initial;               /* Sending initial contents of table? */
     const struct ovsdb_jsonrpc_monitor *monitor;
-    struct json *json;          /* JSON for the whole transaction. */
-
-    /* Current table.  */
     struct ovsdb_jsonrpc_monitor_table *mt;
-    struct json *table_json;    /* JSON for table's transaction. */
 };
 
-static bool
-any_reportable_change(const struct ovsdb_jsonrpc_monitor_table *mt,
-                      const unsigned long int *changed)
+/* Finds and returns the ovsdb_jsonrpc_monitor_row in 'mt->changes' for the
+ * given 'uuid', or NULL if there is no such row. */
+static struct ovsdb_jsonrpc_monitor_row *
+ovsdb_jsonrpc_monitor_row_find(const struct ovsdb_jsonrpc_monitor_table *mt,
+                               const struct uuid *uuid)
+{
+    struct ovsdb_jsonrpc_monitor_row *row;
+
+    HMAP_FOR_EACH_WITH_HASH (row, hmap_node, uuid_hash(uuid), &mt->changes) {
+        if (uuid_equals(uuid, &row->uuid)) {
+            return row;
+        }
+    }
+    return NULL;
+}
+
+/* Allocates an array of 'mt->n_columns' ovsdb_datums and initializes them as
+ * copies of the data in 'row' drawn from the columns represented by
+ * mt->columns[].  Returns the array.
+ *
+ * If 'row' is NULL, returns NULL. */
+static struct ovsdb_datum *
+clone_monitor_row_data(const struct ovsdb_jsonrpc_monitor_table *mt,
+                       const struct ovsdb_row *row)
+{
+    struct ovsdb_datum *data;
+    size_t i;
+
+    if (!row) {
+        return NULL;
+    }
+
+    data = xmalloc(mt->n_columns * sizeof *data);
+    for (i = 0; i < mt->n_columns; i++) {
+        const struct ovsdb_column *c = mt->columns[i].column;
+        const struct ovsdb_datum *src = &row->fields[c->index];
+        struct ovsdb_datum *dst = &data[i];
+        const struct ovsdb_type *type = &c->type;
+
+        ovsdb_datum_clone(dst, src, type);
+    }
+    return data;
+}
+
+/* Replaces the mt->n_columns ovsdb_datums in row[] by copies of the data from
+ * in 'row' drawn from the columns represented by mt->columns[]. */
+static void
+update_monitor_row_data(const struct ovsdb_jsonrpc_monitor_table *mt,
+                        const struct ovsdb_row *row,
+                        struct ovsdb_datum *data)
 {
     size_t i;
 
     for (i = 0; i < mt->n_columns; i++) {
-        const struct ovsdb_jsonrpc_monitor_column *c = &mt->columns[i];
-        unsigned int idx = c->column->index;
+        const struct ovsdb_column *c = mt->columns[i].column;
+        const struct ovsdb_datum *src = &row->fields[c->index];
+        struct ovsdb_datum *dst = &data[i];
+        const struct ovsdb_type *type = &c->type;
 
-        if (c->select & OJMS_MODIFY && bitmap_is_set(changed, idx)) {
-            return true;
+        if (!ovsdb_datum_equals(src, dst, type)) {
+            ovsdb_datum_destroy(dst, type);
+            ovsdb_datum_clone(dst, src, type);
         }
     }
+}
 
-    return false;
+/* Frees all of the mt->n_columns ovsdb_datums in data[], using the types taken
+ * from mt->columns[], plus 'data' itself. */
+static void
+free_monitor_row_data(const struct ovsdb_jsonrpc_monitor_table *mt,
+                      struct ovsdb_datum *data)
+{
+    if (data) {
+        size_t i;
+
+        for (i = 0; i < mt->n_columns; i++) {
+            const struct ovsdb_column *c = mt->columns[i].column;
+
+            ovsdb_datum_destroy(&data[i], &c->type);
+        }
+        free(data);
+    }
+}
+
+/* Frees 'row', which must have been created from 'mt'. */
+static void
+ovsdb_jsonrpc_monitor_row_destroy(const struct ovsdb_jsonrpc_monitor_table *mt,
+                                  struct ovsdb_jsonrpc_monitor_row *row)
+{
+    if (row) {
+        free_monitor_row_data(mt, row->old);
+        free_monitor_row_data(mt, row->new);
+        free(row);
+    }
 }
 
 static bool
 ovsdb_jsonrpc_monitor_change_cb(const struct ovsdb_row *old,
                                 const struct ovsdb_row *new,
-                                const unsigned long int *changed,
+                                const unsigned long int *changed OVS_UNUSED,
                                 void *aux_)
 {
     struct ovsdb_jsonrpc_monitor_aux *aux = aux_;
     const struct ovsdb_jsonrpc_monitor *m = aux->monitor;
     struct ovsdb_table *table = new ? new->table : old->table;
-    enum ovsdb_jsonrpc_monitor_selection type;
-    struct json *old_json, *new_json;
-    struct json *row_json;
-    char uuid[UUID_LEN + 1];
-    size_t i;
+    const struct uuid *uuid = ovsdb_row_get_uuid(new ? new : old);
+    struct ovsdb_jsonrpc_monitor_row *change;
+    struct ovsdb_jsonrpc_monitor_table *mt;
 
     if (!aux->mt || table != aux->mt->table) {
         aux->mt = shash_find_data(&m->tables, table->schema->name);
-        aux->table_json = NULL;
         if (!aux->mt) {
             /* We don't care about rows in this table at all.  Tell the caller
              * to skip it.  */
             return false;
         }
     }
+    mt = aux->mt;
 
-    type = (aux->initial ? OJMS_INITIAL
-            : !old ? OJMS_INSERT
-            : !new ? OJMS_DELETE
-            : OJMS_MODIFY);
-    if (!(aux->mt->select & type)) {
-        /* We don't care about this type of change (but do want to be called
-         * back for changes to other rows in the same table). */
-        return true;
-    }
+    change = ovsdb_jsonrpc_monitor_row_find(mt, uuid);
+    if (!change) {
+        change = xmalloc(sizeof *change);
+        hmap_insert(&mt->changes, &change->hmap_node, uuid_hash(uuid));
+        change->uuid = *uuid;
+        change->old = clone_monitor_row_data(mt, old);
+        change->new = clone_monitor_row_data(mt, new);
+    } else {
+        if (new) {
+            update_monitor_row_data(mt, new, change->new);
+        } else {
+            free_monitor_row_data(mt, change->new);
+            change->new = NULL;
 
-    if (type == OJMS_MODIFY && !any_reportable_change(aux->mt, changed)) {
-        /* Nothing of interest changed. */
-        return true;
-    }
-
-    old_json = new_json = NULL;
-    if (type & (OJMS_DELETE | OJMS_MODIFY)) {
-        old_json = json_object_create();
-    }
-    if (type & (OJMS_INITIAL | OJMS_INSERT | OJMS_MODIFY)) {
-        new_json = json_object_create();
-    }
-    for (i = 0; i < aux->mt->n_columns; i++) {
-        const struct ovsdb_jsonrpc_monitor_column *c = &aux->mt->columns[i];
-        const struct ovsdb_column *column = c->column;
-        unsigned int idx = c->column->index;
-
-        if (!(type & c->select)) {
-            /* We don't care about this type of change for this particular
-             * column (but we will care about it for some other column). */
-            continue;
-        }
-
-        if ((type == OJMS_MODIFY && bitmap_is_set(changed, idx))
-            || type == OJMS_DELETE) {
-            json_object_put(old_json, column->name,
-                            ovsdb_datum_to_json(&old->fields[idx],
-                                                &column->type));
-        }
-        if (type & (OJMS_INITIAL | OJMS_INSERT | OJMS_MODIFY)) {
-            json_object_put(new_json, column->name,
-                            ovsdb_datum_to_json(&new->fields[idx],
-                                                &column->type));
+            if (!change->old) {
+                /* This row was added then deleted.  Forget about it. */
+                hmap_remove(&mt->changes, &change->hmap_node);
+                free(change);
+            }
         }
     }
-
-    /* Create JSON object for transaction overall. */
-    if (!aux->json) {
-        aux->json = json_object_create();
-    }
-
-    /* Create JSON object for transaction on this table. */
-    if (!aux->table_json) {
-        aux->table_json = json_object_create();
-        json_object_put(aux->json, aux->mt->table->schema->name,
-                        aux->table_json);
-    }
-
-    /* Create JSON object for transaction on this row. */
-    row_json = json_object_create();
-    if (old_json) {
-        json_object_put(row_json, "old", old_json);
-    }
-    if (new_json) {
-        json_object_put(row_json, "new", new_json);
-    }
-
-    /* Add JSON row to JSON table. */
-    snprintf(uuid, sizeof uuid,
-             UUID_FMT, UUID_ARGS(ovsdb_row_get_uuid(new ? new : old)));
-    json_object_put(aux->table_json, uuid, row_json);
-
     return true;
 }
 
-/* Returns an overestimate of the number of bytes of JSON data required to
- * report the current contents of the database over monitor 'm'. */
-static size_t
-ovsdb_jsonrpc_monitor_json_length(const struct ovsdb_jsonrpc_monitor *m)
+/* Returns JSON for a <row-update> (as described in RFC 7047) for 'row' within
+ * 'mt', or NULL if no row update should be sent.
+ *
+ * The caller should specify 'initial' as true if the returned JSON is going to
+ * be used as part of the initial reply to a "monitor" request, false if it is
+ * going to be used as part of an "update" notification.
+ *
+ * 'changed' must be a scratch buffer for internal use that is at least
+ * bitmap_n_bytes(mt->n_columns) bytes long. */
+static struct json *
+ovsdb_jsonrpc_monitor_compose_row_update(
+    const struct ovsdb_jsonrpc_monitor_table *mt,
+    const struct ovsdb_jsonrpc_monitor_row *row,
+    bool initial, unsigned long int *changed)
 {
-    const struct shash_node *node;
-    size_t length;
+    enum ovsdb_jsonrpc_monitor_selection type;
+    struct json *old_json, *new_json;
+    struct json *row_json;
+    size_t i;
 
-    /* Top-level overhead of monitor JSON. */
-    length = 256;
+    type = (initial ? OJMS_INITIAL
+            : !row->old ? OJMS_INSERT
+            : !row->new ? OJMS_DELETE
+            : OJMS_MODIFY);
+    if (!(mt->select & type)) {
+        return NULL;
+    }
 
-    SHASH_FOR_EACH (node, &m->tables) {
-        const struct ovsdb_jsonrpc_monitor_table *mt = node->data;
-        const struct ovsdb_table *table = mt->table;
-        const struct ovsdb_row *row;
-        size_t i;
+    if (type == OJMS_MODIFY) {
+        size_t n_changes;
 
-        /* Per-table JSON overhead: "<table>":{...}. */
-        length += strlen(table->schema->name) + 32;
-
-        /* Per-row JSON overhead: ,"<uuid>":{"old":{...},"new":{...}} */
-        length += hmap_count(&table->rows) * (UUID_LEN + 32);
-
-        /* Per-row, per-column JSON overhead: ,"<column>": */
+        n_changes = 0;
+        memset(changed, 0, bitmap_n_bytes(mt->n_columns));
         for (i = 0; i < mt->n_columns; i++) {
-            const struct ovsdb_jsonrpc_monitor_column *c = &mt->columns[i];
-            const struct ovsdb_column *column = c->column;
+            const struct ovsdb_column *c = mt->columns[i].column;
+            if (!ovsdb_datum_equals(&row->old[i], &row->new[i], &c->type)) {
+                bitmap_set1(changed, i);
+                n_changes++;
+            }
+        }
+        if (!n_changes) {
+            /* No actual changes: presumably a row changed and then
+             * changed back later. */
+            return NULL;
+        }
+    }
 
-            length += hmap_count(&table->rows) * (8 + strlen(column->name));
+    row_json = json_object_create();
+    old_json = new_json = NULL;
+    if (type & (OJMS_DELETE | OJMS_MODIFY)) {
+        old_json = json_object_create();
+        json_object_put(row_json, "old", old_json);
+    }
+    if (type & (OJMS_INITIAL | OJMS_INSERT | OJMS_MODIFY)) {
+        new_json = json_object_create();
+        json_object_put(row_json, "new", new_json);
+    }
+    for (i = 0; i < mt->n_columns; i++) {
+        const struct ovsdb_jsonrpc_monitor_column *c = &mt->columns[i];
+
+        if (!(type & c->select)) {
+            /* We don't care about this type of change for this
+             * particular column (but we will care about it for some
+             * other column). */
+            continue;
         }
 
-        /* Data. */
-        HMAP_FOR_EACH (row, hmap_node, &table->rows) {
-            for (i = 0; i < mt->n_columns; i++) {
-                const struct ovsdb_jsonrpc_monitor_column *c = &mt->columns[i];
-                const struct ovsdb_column *column = c->column;
+        if ((type == OJMS_MODIFY && bitmap_is_set(changed, i))
+            || type == OJMS_DELETE) {
+            json_object_put(old_json, c->column->name,
+                            ovsdb_datum_to_json(&row->old[i],
+                                                &c->column->type));
+        }
+        if (type & (OJMS_INITIAL | OJMS_INSERT | OJMS_MODIFY)) {
+            json_object_put(new_json, c->column->name,
+                            ovsdb_datum_to_json(&row->new[i],
+                                                &c->column->type));
+        }
+    }
 
-                length += ovsdb_datum_json_length(&row->fields[column->index],
-                                                  &column->type);
+    return row_json;
+}
+
+/* Constructs and returns JSON for a <table-updates> object (as described in
+ * RFC 7047) for all the outstanding changes within 'monitor', and deletes all
+ * the outstanding changes from 'monitor'.  Returns NULL if no update needs to
+ * be sent.
+ *
+ * The caller should specify 'initial' as true if the returned JSON is going to
+ * be used as part of the initial reply to a "monitor" request, false if it is
+ * going to be used as part of an "update" notification. */
+static struct json *
+ovsdb_jsonrpc_monitor_compose_table_update(
+    const struct ovsdb_jsonrpc_monitor *monitor, bool initial)
+{
+    struct shash_node *node;
+    unsigned long int *changed;
+    struct json *json;
+    size_t max_columns;
+
+    max_columns = 0;
+    SHASH_FOR_EACH (node, &monitor->tables) {
+        struct ovsdb_jsonrpc_monitor_table *mt = node->data;
+
+        max_columns = MAX(max_columns, mt->n_columns);
+    }
+    changed = xmalloc(bitmap_n_bytes(max_columns));
+
+    json = NULL;
+    SHASH_FOR_EACH (node, &monitor->tables) {
+        struct ovsdb_jsonrpc_monitor_table *mt = node->data;
+        struct ovsdb_jsonrpc_monitor_row *row, *next;
+        struct json *table_json = NULL;
+
+        HMAP_FOR_EACH_SAFE (row, next, hmap_node, &mt->changes) {
+            struct json *row_json;
+
+            row_json = ovsdb_jsonrpc_monitor_compose_row_update(
+                mt, row, initial, changed);
+            if (row_json) {
+                char uuid[UUID_LEN + 1];
+
+                /* Create JSON object for transaction overall. */
+                if (!json) {
+                    json = json_object_create();
+                }
+
+                /* Create JSON object for transaction on this table. */
+                if (!table_json) {
+                    table_json = json_object_create();
+                    json_object_put(json, mt->table->schema->name, table_json);
+                }
+
+                /* Add JSON row to JSON table. */
+                snprintf(uuid, sizeof uuid, UUID_FMT, UUID_ARGS(&row->uuid));
+                json_object_put(table_json, uuid, row_json);
+            }
+
+            hmap_remove(&mt->changes, &row->hmap_node);
+            ovsdb_jsonrpc_monitor_row_destroy(mt, row);
+        }
+    }
+
+    free(changed);
+
+    return json;
+}
+
+static bool
+ovsdb_jsonrpc_monitor_needs_flush(struct ovsdb_jsonrpc_session *s)
+{
+    struct ovsdb_jsonrpc_monitor *m;
+
+    HMAP_FOR_EACH (m, node, &s->monitors) {
+        struct shash_node *node;
+
+        SHASH_FOR_EACH (node, &m->tables) {
+            struct ovsdb_jsonrpc_monitor_table *mt = node->data;
+
+            if (!hmap_is_empty(&mt->changes)) {
+                return true;
             }
         }
     }
 
-    return length;
+    return false;
+}
+
+static void
+ovsdb_jsonrpc_monitor_flush_all(struct ovsdb_jsonrpc_session *s)
+{
+    struct ovsdb_jsonrpc_monitor *m;
+
+    HMAP_FOR_EACH (m, node, &s->monitors) {
+        struct json *json;
+
+        json = ovsdb_jsonrpc_monitor_compose_table_update(m, false);
+        if (json) {
+            struct jsonrpc_msg *msg;
+            struct json *params;
+
+            params = json_array_create_2(json_clone(m->monitor_id), json);
+            msg = jsonrpc_create_notify("update", params);
+            jsonrpc_session_send(s->js, msg);
+        }
+    }
 }
 
 static void
 ovsdb_jsonrpc_monitor_init_aux(struct ovsdb_jsonrpc_monitor_aux *aux,
-                               const struct ovsdb_jsonrpc_monitor *m,
-                               bool initial)
+                               const struct ovsdb_jsonrpc_monitor *m)
 {
-    aux->initial = initial;
     aux->monitor = m;
-    aux->json = NULL;
     aux->mt = NULL;
-    aux->table_json = NULL;
 }
 
 static struct ovsdb_error *
@@ -1584,17 +1707,8 @@ ovsdb_jsonrpc_monitor_commit(struct ovsdb_replica *replica,
     struct ovsdb_jsonrpc_monitor *m = ovsdb_jsonrpc_monitor_cast(replica);
     struct ovsdb_jsonrpc_monitor_aux aux;
 
-    ovsdb_jsonrpc_monitor_init_aux(&aux, m, false);
+    ovsdb_jsonrpc_monitor_init_aux(&aux, m);
     ovsdb_txn_for_each_change(txn, ovsdb_jsonrpc_monitor_change_cb, &aux);
-    if (aux.json) {
-        struct jsonrpc_msg *msg;
-        struct json *params;
-
-        params = json_array_create_2(json_clone(aux.monitor->monitor_id),
-                                     aux.json);
-        msg = jsonrpc_create_notify("update", params);
-        jsonrpc_session_send(aux.monitor->session->js, msg);
-    }
 
     return NULL;
 }
@@ -1604,8 +1718,9 @@ ovsdb_jsonrpc_monitor_get_initial(const struct ovsdb_jsonrpc_monitor *m)
 {
     struct ovsdb_jsonrpc_monitor_aux aux;
     struct shash_node *node;
+    struct json *json;
 
-    ovsdb_jsonrpc_monitor_init_aux(&aux, m, true);
+    ovsdb_jsonrpc_monitor_init_aux(&aux, m);
     SHASH_FOR_EACH (node, &m->tables) {
         struct ovsdb_jsonrpc_monitor_table *mt = node->data;
 
@@ -1617,7 +1732,8 @@ ovsdb_jsonrpc_monitor_get_initial(const struct ovsdb_jsonrpc_monitor *m)
             }
         }
     }
-    return aux.json ? aux.json : json_object_create();
+    json = ovsdb_jsonrpc_monitor_compose_table_update(m, true);
+    return json ? json : json_object_create();
 }
 
 static void
@@ -1629,6 +1745,14 @@ ovsdb_jsonrpc_monitor_destroy(struct ovsdb_replica *replica)
     json_destroy(m->monitor_id);
     SHASH_FOR_EACH (node, &m->tables) {
         struct ovsdb_jsonrpc_monitor_table *mt = node->data;
+        struct ovsdb_jsonrpc_monitor_row *row, *next;
+
+        HMAP_FOR_EACH_SAFE (row, next, hmap_node, &mt->changes) {
+            hmap_remove(&mt->changes, &row->hmap_node);
+            ovsdb_jsonrpc_monitor_row_destroy(mt, row);
+        }
+        hmap_destroy(&mt->changes);
+
         free(mt->columns);
         free(mt);
     }
